from contextlib import contextmanager
import functools

import select
import socket

from . import transport, transceiver, pool

def can_accept(serv_sock):
	return serv_sock in select.select([serv_sock], [], [], 0)[0]

@contextmanager
def clean_exit(cleanup, *ignore_exc):
	try:
		yield None
	except BaseException as exc:
		# catch BaseException to include
		# non-error Exceptions such as KeyboardInterrupt
		if not any([isinstance(exc, exc_type) for exc_type in ignore_exc]):
			raise
	finally:
		cleanup()

def close(*args):
	for arg in args: arg.close()

class sockmanager:
	def __init__(self):
		self._socks = []

	def __iter__(self):
		self.update()
		return iter(self._socks)

	def update(self):
		self._socks[:] = [sock for sock in self._socks if sock.open or sock.receiver.pending > 0]

	def put(self, sock):
		self._socks.append(sock)

def main(port=7001):
	pools = pool.PoolManager(False)
	serv = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	serv.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
	serv.bind(('0.0.0.0', port))
	serv.setblocking(0)
	serv.listen(5)

	mgr = sockmanager()

	import time
	with clean_exit(functools.partial(close, serv, pools), KeyboardInterrupt):
		while True:
			while can_accept(serv):
				conn, addr = serv.accept()
				trans = transceiver.WebsocketTransceiver.of(conn)
				wsock = trans.socket
				wsock.receiver = transport.EventReceiver()
				wsock.transceiver = trans
				wsock.add_receiver(wsock.receiver)
				mgr.put(wsock)
				pools.add_endpoint(trans)

			time.sleep(.01)
			pools.update()

			for sock in mgr:
				for name, data in sock.receiver:
					if name == "describe":
						sp = pools.get_pool_for(sock.transceiver)
						sp.describe(sock.transceiver, data)

if __name__ == '__main__':
	main()


from abc import ABC, abstractmethod

from . import transport

class Transceiver(ABC):
	def __init__(self):
		self.id = None

	@abstractmethod
	def is_open(self):
		"""
	`	A method used to check whether the channel
		used to transmit and received data is open.
		"""
		return True

	@abstractmethod
	def send(self, event_type, event_data):
		"""
		A method which sends an event to an endpoint. How the
		data is sent or formatted is defined by individual
		subclasses and is not important as long as it sent somehow.

		:param event_type: The type of event to be processed.
		:type event: str
		:param event_data: Data describing the event.
		:type event_data: dict
		"""
		pass

	@abstractmethod
	def receive(self):
		"""
		A generator yielding all relevant events at the time it is called.

		Events should be yielded as a 2-tuple (event_type, event_data)
		"""
		pass

	@abstractmethod
	def close(self):
		pass

class WebsocketTransceiver(Transceiver):
	def __init__(self, ws):
		super().__init__()
		if not isinstance(ws, transport.Websocket):
			raise TypeError("ws must be a Websocket instance!")
		else:
			self._ws = ws
			self._recv = transport.EventReceiver()
			self._ws.add_receiver(self._recv)

	def is_open(self):
		return self._ws.open

	def send(self, event_type, event_data):
		self._ws.send(transport.create_event(event_type, event_data))

	def receive(self):
		self._ws.receive()
		yield from self._recv

	def close(self):
		self._ws.close_raw()

	@property
	def socket(self):
		return self._ws

	@staticmethod
	def of(raw_sock):
		porter = transport.Websocket(raw_sock)
		return WebsocketTransceiver(porter)

class TransceiverManager:
	def __init__(self):
		self._trans = []

	def dispatch(self, transceiver):
		self._trans.append(transceiver)

	def events(self):
		"""
		Used by PoolManager to receive all events from all
		transceivers and handle them.
		"""
		for trans in self._trans:
			for event in trans.receive():
				yield (event, trans)

		self._trans[:] = [trans for trans in self._trans if trans.is_open()]

	def close_all(self):
		for trans in self._trans:
			trans.close()

		self._trans = []


import pathlib
from setuptools import setup, find_packages

__dirname__ = pathlib.Path(__file__).parent
readme_text = (__dirname__ / "README.md").read_text()

setup(
	name="pyrtc",
	version="1.0.0",
	description="Python WebRTC signalling and connection management back-end.",
	long_description=readme_text,
	long_description_content_type="text/markdown",
	# TODO: url=
	author="Isaac Dorenkamp",
	author_email="unorthodox.isaac@gmail.com",
	license="MIT",
	classifiers=[
		"License :: OSI Approved :: MIT License",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.9"
	],
	package_dir={ "": "src" },
	packages=find_packages(where="src"),
	include_package_data=True,
	install_requires=["h11==0.13.0", "wsproto==1.0.0"]
)


import json

import select
import socket
import ssl

from wsproto import events, ConnectionType, WSConnection
from wsproto.connection import ConnectionState

def can_read(sock):
	if isinstance(sock, ssl.SSLSocket):
		return sock.pending() > 0
	else:
		return sock in select.select([sock], [], [], 0)[0]

def can_write(sock):
	# should work for both SSL and non-SSL sockets
	return sock in select.select([], [sock], [], 0)[1]

def create_event(name, data):
	return events.TextMessage(json.dumps({
		'type': name,
		'data': data
	}))

class EventReceiver:
	def __init__(self):
		self._buffer = []

	def __iter__(self):
		return self

	def __next__(self):
		if len(self._buffer) > 0:
			return self._buffer.pop(0)
		else:
			raise StopIteration

	def post(self, evt):
		self._buffer.append(evt)

	@property
	def pending(self):
		return len(self._buffer)

class Websocket:
	def __init__(self, cli_sock):
		self._sock = cli_sock
		self._conn = WSConnection(ConnectionType.SERVER)
		self._closed = False
		self._receivers = []

	def add_receiver(self, receiver):
		if isinstance(receiver, EventReceiver):
			self._receivers.append(receiver)
		else:
			raise TypeError("receiver must be an instance of EventReceiver")

	def _raw_send(self, data):
		if self._sock.fileno() == -1:
			self._closed = True
			return

		sent = 0
		while sent < len(data):
			if can_write(self._sock):
				try:
					sent += self._sock.send(data[sent:])
				except socket.error:
					# broken pipe
					self._closed = True
					return
			else:
				time.sleep(0.05)

	@property
	def open(self):
		return self._conn.state == ConnectionState.OPEN and not self._closed

	def close(self, status=1000):
		if self._conn.state != ConnectionState.OPEN:
			raise ValueError("WebSocket not in the OPEN state.")
		else:
			close = self._conn.send(events.CloseConnection(status))
			self._raw_send(close)
			self._sock.close()

	def close_raw(self):
		if self._conn.state == ConnectionState.OPEN:
			self.close()
		else:
			self._closed = True
			self._sock.close()

	def send(self, data):
		if isinstance(data, str):
			message = events.TextMessage(data)
		elif isinstance(data, bytes):
			message = events.BytesMessage(data)
		elif isinstance(data, (events.TextMessage, events.BytesMessage)):
			message = data
		else:
			raise TypeError("message must be a str or bytes object, or a TextMessage/BytesMessage")

		raw_data = self._conn.send(message)
		self._raw_send(raw_data)

	def receive(self):
		if self._closed:
			return

		while can_read(self._sock):
			data = self._sock.recv(2048)
			if not data:
				# socket closed
				self._closed = True
				return
			self._conn.receive_data(data)

		for event in self._handle(self._conn.events()):
			for recv in self._receivers:
				recv.post(event)

	def _handle(self, event_iter):
		for event in event_iter:
			if isinstance(event, events.Request):
				self._raw_send(self._conn.send(events.AcceptConnection()))
			elif isinstance(event, events.TextMessage):
				if event.message_finished:
					try:
						evt_data = json.loads(event.data)
					except json.decoder.JSONDecodeError:
						# tell client to send valid data!
						self._raw_send(self._conn.send(create_event('rtc:error', { 'message': 'Please send valid JSON data.' })))

					name = evt_data.get('type', None)
					data = evt_data.get('data', None)
					if not (name and isinstance(data, dict)):
						self._raw_send(self._conn.send(create_event('rtc:error', { 'message': 'Please send an event with "type" and "data" fields.' })))
					else:
						yield (name, data)
			elif isinstance(event, events.Ping):
				# send pong, nothing to yield
				self._raw_send(self._conn.send(event.response()))
			elif isinstance(event, events.CloseConnection):
				self.close_raw()
				return
			else:
				# ignore all else
				pass


import uuid

from .transceiver import Transceiver, TransceiverManager

class Pool:
	def __init__(self, name, timeout=30):
		"""
		Create a pool with the given unique name and timeout.
		timeout is the number of seconds after which a pool will
		automatically close without activity.
		"""
		self._name = name
		self._conns = {}
		self._closed = False
		self.timeout = timeout

	def get_peer_ids(self):
		"""
		Returns a tuple of the IDs of all connected peers.
		"""
		return tuple(self._conns.keys())

	@property
	def name(self):
		return self._name

	@property
	def closed(self):
		return self._closed

	def allows(self, endpoint):
		"""
		A method which returns a boolean
		value telling whether the given
		endpoint is allowed to join this
		pool. Subclasses may return False
		in some cases, blocking certain
		endpoints from joining.
		"""
		return not self.closed

	def get_connection(self, test):
		"""
		If test is a string, gets a connection by its ID.
		If test is a Transceiver, gets the connection associated with it.
		If no connection is found, None is returned.
		"""
		if isinstance(test, str):
			return self._conns.get(test, None)
		elif isinstance(test, Transceiver):
			for i in self._conns.values():
				if i['socket'] == test:
					return i
			return None
		else:
			raise TypeError("get_connection accepts a UUID string or a Transceiver.")

	def add_endpoint(self, transceiver):
		"""
		Accepts a Transceiver and adds it to the pool
		as an endpoint.
		"""
		if transceiver in self:
			raise ValueError("This endpoint has already been added!")

		ent = {
			'socket': transceiver,
			'id': str(uuid.uuid4()),
			'description': None
		}

		self._conns[ent['id']] = ent
		return ent

	def update(self):
		"""
		Prunes closed connections and notifies peers of the closure.
		"""
		if self.closed: return

		# remove closed transceivers,
		# send close notifications to peers
		pre_load = tuple(self._conns.values())
		closed = []
		for conn in pre_load:
			if not conn['socket'].is_open():
				closed.append(conn['id'])
				del self._conns[conn['id']]

		for c in closed:
			self.broadcast('rtc:close', { 'uid': c })

	def describe(self, id_or_trans, desc):
		"""
		Sets the description of the specified connection.
		"""
		if self.closed:
			raise ValueError("Pool is closed")

		assert desc is None or isinstance(desc, dict)
		conn = self.get_connection(id_or_trans)
		if conn:
			self.broadcast('rtc:describe', {
				'description': desc,
				'uid': conn['id']
			})

	def broadcast(self, evt_type, evt_data, *exclude):
		"""
		Broadcasts the specified event to all connections in the pool,
		except for those whose Transceiver is in ``exclude``.
		"""
		if self.closed:
			raise ValueError("Pool is closed")

		for conn in filter(lambda ent: ent['socket'] not in exclude, self._conns.values()):
			conn['socket'].send(evt_type, evt_data)

	def close(self):
		"""
		Sends a signal to close all connections in the pool,
		and closes the signalling channels.
		"""

		if self.closed:
			raise ValueError("Pool is closed")

		self.broadcast('rtc:stop', {})
		for conn in self._conns:
			conn['socket'].close()

		self._conns[:] = []
		self._closed = True

	def __contains__(self, test):
		if isinstance(test, str):
			# check by UID
			return test in self._conns.keys()
		elif isinstance(test, Transceiver):
			return test in map(lambda ent: ent['socket'], self._conns.values())
		else:
			return False

class PoolManager:
	def __init__(self, private=True):
		self._private = private
		self._pools = []
		self._trans = TransceiverManager()

		self._closed = False

	def close(self):
		"""
		Closes this pool manager.
		"""
		self._trans.close_all()
		self._closed = True

	def _close_assert(self, msg="PoolManager is closed"):
		assert not self._closed, msg

	def update(self):
		self._close_assert()
		for event, source in self._trans.events():
			self.handle_event(event, source)

		immut = tuple(self._pools)
		for pool in immut:
			if pool.closed:
				self._pools.remove(pool)
			else:
				pool.update()

	def add_endpoint(self, transceiver):
		self._close_assert()
		self._trans.dispatch(transceiver)

	def add_pool(self, pool):
		self._close_assert()
		if pool.name in map(lambda pool: pool.name, self._pools):
			raise ValueError(f"Pool with name '{pool.name}' already exists!")
		else:
			self._pools.append(pool)

	def get_pool(self, name):
		for pool in self._pools:
			if pool.name == name: return pool

		return None

	def remove_pool(self, name):
		self._close_assert()
		pool = self.get_pool(name)
		if pool is not None:
			self._pools.remove(pool)
			return True
		else:
			return False

	def _query(self, test):
		for pool in self._pools:
			conn = pool.get_connection(test)
			if conn:
				return pool, conn

		return None, None

	def get_pool_for(self, test):
		return self._query(test)[0]

	def get_connection_for(self, test):
		return self._query(test)[1]

	def handle_event(self, event, source):
		"""
		:param: event
		:type event: tuple
		:param source:
		:type source: Transceiver
		"""

		self._close_assert()

		name, data = event
		parts = name.split(':')
		prefix = parts[0]
		if prefix != 'rtc' or len(parts) != 2:
			# Ignore events not intended for the rtc
			# signalling process. This allows a single
			# websocket to be used for RTC signalling
			# as well as other applications.
			return

		action = parts[1]
		if action == 'join':
			_pool = data.get('pool', None)
			if not _pool:
				source.send('rtc:error', {
					'message': 'Please specify a pool to join.'
				})
			pool = self.get_pool(_pool)

			if pool:
				can_join = pool.allows(source)
				if can_join:
					peers = pool.get_peer_ids()
					result = pool.add_endpoint(source)
					source.id = result['id']
					joined = ('rtc:joined', {
						'client_id': result['id'],
						'peers': peers,
						'pool': pool.name
					})
				else:
					source.send('rtc:error', {
						'message': 'You are not authorized to join.',
						'pool': pool.name
					})
					return
			else:
				if self._private:
					# pools cannot be created implicitly,
					# must be explicitly created by server code
					source.send('rtc:error', {
						'message': 'No such pool exists.',
						'pool': _pool
					})
					return
				else:
					# auto-add default pool
					pool = Pool(_pool)
					self.add_pool(pool)
					peers = pool.get_peer_ids()
					ent = pool.add_endpoint(source)
					source.id = ent['id']
					joined = ('rtc:joined', {
						'client_id': ent['id'],
						'peers': peers,
						'pool': pool.name
					})

			# if we get here, joined will have been assigned
			source.send(*joined)
			pool.broadcast("rtc:request_offers", {
				"for": source.id
			}, source)
		elif action == 'offers':
			for key in data.keys():
				pool, conn = self._query(key)
				if conn:
					conn['socket'].send('rtc:offer', {
						'for': source.id,
						'offer': data[key]
					})
		elif action == 'answer':
			target = data.get('for', None)
			if target:
				conn = self.get_connection_for(target)
				if conn:
					conn['socket'].send('rtc:answer', {
						'for': source.id,
						'answer': data.get('answer')
					})
			else:
				source.send('rtc:error', {
					'message': 'Please specify an answer target.'
				})
		elif action == 'candidate':
			target = data.get('for', None)
			if target:
				conn = self.get_connection_for(target)
				if conn:
					conn['socket'].send('rtc:candidate', {
						'candidate': data.get('candidate'),
						'from': source.id
					})
			else:
				source.send('rtc:error', {
					'message': "Please specify a candidate recipient in the 'for' field."
				})
		else:
			source.send('rtc:error', {
				'message': f"RTC action '{action}' not recognized."
			})